﻿local Links = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceHook-2.1", "AceDB-2.0")

local L = AceLibrary("AceLocale-2.2"):new("Links")
local dewdrop = AceLibrary("Dewdrop-2.0")
local event

L:RegisterTranslations("enUS", function() return {
	["Link in chat"] = true,
	["Dressup"] = true,
	["Close all Tooltips"] = true,
	["Links - plenty of them"] = true,
	["modinvert"] = true,
	["Invert the alt-clic modifier behavior of Links."] = true,
	["autohide"] = true,
	["Automatically hide tooltip on escape."] = true,
	["autocomp"] = true,
	["Automatically show the compare window when hovering tooltips."] = true,
	["showicon"] = true,
	["Show item icon with the tooltip."] = true,
	["CHAT_COMMANDS"] = { "/links" },
} end)

L:RegisterTranslations("koKR", function() return {
	["Link in chat"] = "대화창에 링크",
	["Dressup"] = "미리보기",
	["Close all Tooltips"] = "모든 툴팁 닫기",
	["Links - plenty of them"] = "링크 - 많은 아이템 링크 확인",
	["modinvert"] = "모드 전환",
	["Invert the alt-clic modifier behavior of Links."] = "ALT+클릭으로 링크 확인을 하도록 변경합니다.",
	["autohide"] = "자동 숨김",
	["Automatically hide tooltip on escape."] = "Esc키 입력시 툴팁을 숨김니다.",
 	["autocomp"] = "자동 비교",
 	["Automatically show the compare window when hovering tooltips."] = "툴팁에 마우스 오버시 자동으로 비교 창을 표시합니다.",
	["showicon"] = "아이콘 표시",
	["Show item icon with the tooltip."] = "툴팁에 아이템의 아이콘을 표시합니다.",
} end)

L:RegisterTranslations("frFR", function() return {
	["Link in chat"] = "Insérer un lien dans la fenêtre de discussion",
	["Dressup"] = "Voir l'objet dans la cabine d'essayage",
	["Close all Tooltips"] = "Fermer toutes les fenêtres d'objets",
	["Invert the alt-clic modifier behavior of Links."] = "Inverser le comportement de Links sur le alt-clic.",
	["Automatically hide tooltip on escape."] = "Cacher automatiquement les tooltips lorsque Echap est appuyé.",
	["Automatically show the compare window when hovering tooltips."] = "Affiche automatiquement les fenêtres de comparaison lorsque le curseur est au dessus des tooltips.",
	["Show item icon with the tooltip."] = "Afficher l'icône des objets avec le tooltip.",
} end)

L:RegisterTranslations("zhTW", function() return {
	["Link in chat"] = "在聊天視窗貼上物品連結",
	["Dressup"] = "試穿",
	["Close all Tooltips"] = "關閉全部物品連結提示訊息",
	["Links - plenty of them"] = "Links - 充足的物品連結提示訊息",
	["modinvert"] = "反轉輔助鍵",
	["Invert the alt-clic modifier behavior of Links."] = "反轉 Links 的Alt-點擊行為模式。",
	["autohide"] = "自動隱藏",
	["Automatically hide tooltip on escape."] = "按下Esc鍵自動隱藏提示訊息。",
	["autocomp"] = "自動比較",
	["Automatically show the compare window when hovering tooltips."] = "當游標在物品連結提示訊息上時自動顯示比較視窗。",
	["showicon"] = "顯示圖示",
	["Show item icon with the tooltip."] = "在物品連結提示訊息上顯示圖示。",
	["CHAT_COMMANDS"] = { "/links" },
} end)

Links:RegisterDB("LinksDB")
Links:RegisterDefaults("profile", {
	modinvert = false,
	autohide = true,
	autocomp = false,
	showicon = false,
	positions = {},
})

function Links:SavePosition(frame)
	local index = frame.index
	local position = self.db.profile.positions[index]
	if not position then
		position = {}
		self.db.profile.positions[index] = position
	end
	position.left, position.top = frame:GetLeft(), frame:GetTop()
end

function Links:OnInitialize()
	self.functionNames, self.functionHandlers = {}, {}
	self:AddToMenu(L["Link in chat"],
		function (_, text)
			if not ChatFrameEditBox:IsVisible() then
				ChatFrame_OpenChat (text)
			else
				ChatFrameEditBox:Insert(text)
			end
		end)
	self:AddToMenu(L["Dressup"],
		function (_, _, link)
			DressUpItemLink(link)
		end)
	self:AddToMenu(L["Close all Tooltips"],
		function ()
			Links:HideAllFrames()
		end)
	
	self.frames = {}
	self.index = 1

	self.options = {
		desc = L["Links - plenty of them"],
		type = "group",
		args = {
			modinvert = {
				name = L["modinvert"],
				desc = L["Invert the alt-clic modifier behavior of Links."],
				type = "toggle",
				get = function() return self.db.profile.modinvert end,
				set = function(value) self.db.profile.modinvert = value end,
			},
			autohide = {
				name = L["autohide"],
				desc = L["Automatically hide tooltip on escape."],
				type = "toggle",
				get = function() return self.db.profile.autohide end,
				set = function(value) self:SetAutoHide(value) end,
			},
			autocomp = {
				name = L["autocomp"],
				desc = L["Automatically show the compare window when hovering tooltips."],
				type = "toggle",
				get = function() return self.db.profile.autocomp end,
				set = function(value) self.db.profile.autocomp = value end,
			},
			showicon = {
				name = L["showicon"],
				desc = L["Show item icon with the tooltip."],
				type = "toggle",
				get = function() return self.db.profile.showicon end,
				set = function(value) self.db.profile.showicon = value end,
			}
		},
	}

	self:RegisterChatCommand(L["CHAT_COMMANDS"], self.options)
end

function Links:AddToMenu(name, handler)
	local names, handlers = self.functionNames, self.functionHandlers
	names[#names + 1] = name
	handlers[#handlers + 1] = handler
end

function Links:OnEnable()
	self:Hook("SetItemRef", true)
	local Cartographer_InstanceLoot = Cartographer_InstanceLoot
	if Cartographer_InstanceLoot then
		function self:Cartographer_InstanceLoot_DropClick(_, id)
			local name, link = GetItemInfo(id)
			if not name then
				self.hooks[Cartographer_InstanceLoot].DropClick(Cartographer_InstanceLoot, id)
			elseif not HandleModifiedItemClick(link) then
				self:OpenLink(nil, link)
			end
		end
		self:Hook(Cartographer_InstanceLoot, "DropClick", "Cartographer_InstanceLoot_DropClick")
	end
end

function Links:SetAutoHide(value)
	local UISpecialFrames = UISpecialFrames
	self.db.profile.autohide = value
	for frame in pairs(self.frames) do
		local name = frame:GetName()
		if value then
			UISpecialFrames[#UISpecialFrames + 1] = name
		else
			for i = 1, #UISpecialFrames do
				if name == UISpecialFrames[i] then
					table.remove(UISpecialFrames, i)
					break
				end
			end
		end
	end
end

function Links:OpenMenu(frame)
	dewdrop:Open(frame,
		"point", function() return "TOPLEFT", "TOPRIGHT" end,
		"dontHook", true,
		"children", function (level, name)
			local names, handlers = self.functionNames, self.functionHandlers
			for i = 1, #names do
				local name, handler = names[i], handlers[i]
				dewdrop:AddLine(
					"text", name,
					"func", function ()
						handler(frame, frame.text, frame.link)
						dewdrop:Close()
					end)
			end
		end)
end

function Links:HideAllFrames()
	for frame in pairs(self.frames) do
		if frame:IsVisible() then
			frame:Hide()
		end
	end
end

Links.NORMAL_COLOR = { 0.09, 0.12, 0.09 }
Links.ANCHORED_COLOR = { 0.18, 0.09, 0.09 }

local LinksFrame_Events = {
	OnDragStart = function (self)
		self:StartMoving()
	end,
	OnDragStop = function (self)
		self:StopMovingOrSizing()
		Links:ValidateFramePosition(self)
	end,
	OnHide = function (self)
		local left, top = self:GetLeft(), self:GetTop()
		self:ClearAllPoints()
		self:SetPoint("TOPLEFT", left, top - UIParent:GetHeight())
		self:SetBackdropColor(unpack(Links.NORMAL_COLOR))
		self.link = nil
		self.text = nil
		Links:SavePosition(self)
		Links:OnHideTooltip(self)
	end,
	OnReceiveDrag = function (self)
		local type, id, link = GetCursorInfo()
		if type == "item" then
			Links:OpenLink(nil, link, self)
			ClearCursor()
		end
	end,
	OnEnter = function (self)
		local autocomp = Links.db.profile.autocomp
		if autocomp or IsModifiedClick("COMPAREITEMS") then
			Links:ShowCompareItem(self)
		end
		if not autocomp then
			self:RegisterEvent("MODIFIER_STATE_CHANGED")
		end
	end,
	OnLeave = function (self)
		Links:HideCompareItems(self)
		self:UnregisterEvent("MODIFIER_STATE_CHANGED")
	end,
	OnEvent = function (self, event)
		if event == "MODIFIER_STATE_CHANGED" then
			if IsModifiedClick("COMPAREITEMS") then
				Links:ShowCompareItem(self)
			else
				Links:HideCompareItems(self)
			end
		end
	end,
}

local ShoppingTooltip1, ShoppingTooltip2 = ShoppingTooltip1, ShoppingTooltip2
function Links:ShowCompareItem(frame)
	local _, link = frame:GetItem()
	if not link then return end

	local item2
	if ShoppingTooltip1:SetHyperlinkCompareItem(link, 1) then
		if ShoppingTooltip2:SetHyperlinkCompareItem(link, 2) then
			item2 = true
		end
	else
		return
	end
	local left, right = frame:GetLeft(), frame:GetRight()
	local side
	local screen = GetScreenWidth()
	if (screen - right) < left then
		side = "left"
	else
		side = "right"
	end
	local anchor = frame:GetAnchorType()
	if anchor then
		local width = ShoppingTooltip1:GetWidth() +
			(item2 and ShoppingTooltip2:GetWidth() or 0)

		if side == "left" then
			local delta = width - left
			if delta > 0 then
				frame:SetAnchorType(anchor, delta, 0)
			end
		elseif side == "right" then
			local delta = screen - (right + width)
			if delta < 0 then
				frame:SetAnchorType(anchor, delta, 0)
			end
		end
	end

	ShoppingTooltip1:SetOwner(frame, "ANCHOR_NONE")
	ShoppingTooltip1:ClearAllPoints()
	if side == "left" then
		ShoppingTooltip1:SetPoint("TOPRIGHT", frame, "TOPLEFT", 0, -10)
	else
		ShoppingTooltip1:SetPoint("TOPLEFT", frame, "TOPRIGHT", 0, -10)
	end
	ShoppingTooltip1:SetHyperlinkCompareItem(link, 1)
	ShoppingTooltip1:Show()

	if item2 then
		ShoppingTooltip2:SetOwner(ShoppingTooltip1, "ANCHOR_NONE")
		ShoppingTooltip2:ClearAllPoints()
		if side == "left" then
			ShoppingTooltip2:SetPoint("TOPRIGHT", ShoppingTooltip1, "TOPLEFT", 0, 0)
		else
			ShoppingTooltip2:SetPoint("TOPLEFT", ShoppingTooltip1, "TOPRIGHT", 0, 0)
		end
		ShoppingTooltip2:SetHyperlinkCompareItem(link, 2)
		ShoppingTooltip2:Show()
	end
end

function Links:HideCompareItems(frame)
	if ShoppingTooltip1:IsOwned(frame) then
		ShoppingTooltip1:Hide()
		ShoppingTooltip2:Hide()
	end
end

function Links:MakeFrame()
	local name = string.format("LinksTooltip%d", self.index)
	local frame = CreateFrame("GameTooltip", name, UIParent, "GameTooltipTemplate")
	frame.index = self.index

	self.index = self.index + 1
	self.frames[frame] = true

	local close = CreateFrame("Button", name.."Close", frame)
	frame.close = close

	local makeTexture = function (path, blend)
		local t = close:CreateTexture()
		t:SetAllPoints(close)
		t:SetTexture(path)
		if blend then
			t:SetBlendMode(blend)
		end
		return t
	end

	close:SetNormalTexture(makeTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up"))
	close:SetPushedTexture(makeTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down"))
	close:SetHighlightTexture(makeTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight", "ADD"))

	close:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	close:SetScript("OnClick",
		function (_, button)
			if button == "LeftButton" then
				frame:Hide()
			else
				self:OpenMenu(frame)
			end
		end)


	close:SetWidth(32)
	close:SetHeight(32)
	close:ClearAllPoints()
	close:SetPoint("TOPRIGHT", 1, 0)

	close:Show()

	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetToplevel(true)
	frame:SetFrameStrata("HIGH")
	local position = self.db.profile.positions[frame.index]
	if position then
		frame:SetPoint("TOPLEFT", position.left, position.top - UIParent:GetHeight())
	else
		frame:SetPoint("CENTER")
	end

	frame:SetPadding(16)
	frame:RegisterForDrag("LeftButton")
	frame:SetBackdropColor(unpack(self.NORMAL_COLOR))

	for event, handler in pairs(LinksFrame_Events) do
		frame:SetScript(event, handler)
	end

	if self.db.profile.autohide then
		local UISpecialFrames = UISpecialFrames
		UISpecialFrames[#UISpecialFrames + 1] = name
	end

	local Mendeleev = Mendeleev
	if Mendeleev and Mendeleev.OnTooltipSetItem then
		Mendeleev:HookScript(frame, "OnTooltipSetItem")
		Mendeleev:HookScript(frame, "OnTooltipCleared")
	end

	return frame
end

do
    local ANCHORING_DISTANCE = 10
    local ANCHORING_OFFSET_X = 0
    local abs = math.abs

	local function GetAnchorX(frame, other)
        local fleft, fright = frame:GetLeft(), frame:GetRight()
        local oleft, oright = other:GetLeft(), other:GetRight()

		if abs(fleft - oleft) < ANCHORING_DISTANCE then
			return "LEFT", "LEFT"
		elseif abs(fleft - oright) < ANCHORING_DISTANCE then
			return "LEFT", "RIGHT"
		elseif abs(fright - oright) < ANCHORING_DISTANCE then
			return "RIGHT", "RIGHT"
		elseif abs(fright - oleft) < ANCHORING_DISTANCE then
			return "RIGHT", "LEFT"
		end
	end

	local function GetAnchorY(frame, other)
        local ftop, fbottom = frame:GetTop(), frame:GetBottom()
        local otop, obottom = other:GetTop(), other:GetBottom()

		if abs(ftop - otop) < ANCHORING_DISTANCE then
			return "TOP", "TOP"
		elseif abs(ftop - obottom) < ANCHORING_DISTANCE then
			return "TOP", "BOTTOM"
		elseif abs(fbottom - obottom) < ANCHORING_DISTANCE then
			return "BOTTOM", "BOTTOM"
		elseif abs(fbottom - otop) < ANCHORING_DISTANCE then
			return "BOTTOM", "TOP"
		end
	end

    function Links:IsAnchoring(frame, other)
		local fanchorx, oanchorx = GetAnchorX(frame, other)
		if not fanchorx then return end
		local fanchory, oanchory = GetAnchorY(frame, other)
		if not fanchory then return end

		if (fanchorx == oanchorx) == (fanchory == oanchory) then return end

		return fanchory..fanchorx, oanchory..oanchorx, 0, 0
    end
end

function Links:OnHideTooltip(frame)
    for other in pairs(self.frames) do
		if other.anchored == frame then
			local left, top = other:GetLeft(), other:GetTop()
			other:ClearAllPoints()
			other:SetPoint("TOPLEFT", left, top - UIParent:GetHeight())
			other:SetBackdropColor(unpack(self.NORMAL_COLOR))
			other.anchored = nil
			self:SavePosition(other)
		end
	end
end

function Links:ValidateFramePosition(frame)
    -- check anchoring
    for other in pairs(self.frames) do
        if other ~= frame and other:IsVisible() and not other.anchored then
            local point, relative, x, y = self:IsAnchoring(frame, other)
            if point then
                frame:ClearAllPoints()
                frame:SetBackdropColor(unpack(self.ANCHORED_COLOR))
                frame:SetPoint(point, other, relative, x, y)
				frame.anchored = other
                return
            end
        end
    end
    local left, top = frame:GetLeft(), frame:GetTop()
    frame:ClearAllPoints()
    frame:SetPoint("TOPLEFT", left, top-UIParent:GetHeight())
    frame:SetBackdropColor(unpack(self.NORMAL_COLOR))
	frame.anchored = nil
    ValidateFramePosition(frame)
	self:SavePosition(frame)
end

function Links:GetAvailableFrame()
	for frame in pairs(self.frames) do
		if not frame:IsVisible() then
			return frame
		end
	end
	return self:MakeFrame()
end

function Links:OpenLink(link, text, frame)
	link = link or (text and text:match("|H([^|]+)|h"))
	if not link then return end

	frame = frame or self:GetAvailableFrame()
	frame:ClearLines()
	frame:SetOwner(UIParent, "ANCHOR_PRESERVE")
	frame:SetHyperlink(link)
	frame.link = link
	frame.text = text

	local icon = frame.icon
	if self.db.profile.showicon and link:sub(1, 4) == "item" then
		if not icon then
			icon = frame:CreateTexture(nil, "ARTWORK") 
			frame.icon = icon
			icon:SetHeight(36)
			icon:SetWidth(36)
			icon:ClearAllPoints();
			icon:SetPoint("TOPRIGHT", frame, "TOPLEFT", "1", "-2")
		end
		icon:SetTexture(select(10, GetItemInfo(link)))
		icon:Show()
	elseif icon then
		icon:Hide()
	end

	if not event then
		event = AceLibrary("AceEvent-2.0")
	end
	event:TriggerEvent("Links_LinkOpened", link, frame)
	--[[
		The API is to do something like this in your addon :

		function MyAddon:OnEnable()
			...
			self:RegisterEvent("Links_LinkOpened")
		end

		function MyAddon:Links_LinkOpened(link, frame)
			...
		end
	]]
	frame:Show()
end

function Links:SetItemRef(link, text, button)
	if link:sub(1, 4) ~= "item" and link:sub(1, 7) ~= "enchant" then
		return self.hooks.SetItemRef(link, text, button)
	else
		if not self.db.profile.modinvert == not not IsModifiedClick("Links_Disable") then
			if not HandleModifiedItemClick(text) then
				ShowUIPanel(ItemRefTooltip)
				if not ItemRefTooltip:IsShown() then
					ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
				end
				ItemRefTooltip:SetHyperlink(link)
			end
			return
		end
	end
	if not HandleModifiedItemClick(text) then
		for frame in pairs(self.frames) do
			if frame:IsVisible() and frame.link == link then
				frame:Hide()
				frame.link = nil
				return
			end
		end
		self:OpenLink(link, text)
	end
end

_G.Links = Links
